.custom-widget {
  padding: 0px 2px;
  font-family: system-ui, -apple-system, blinkmacsystemfont, "Segoe UI",
    helvetica, arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji",
    "Segoe UI Symbol";
}

.jp-LinkedOutputView .jp-OutputArea-output:only-child {
  margin: 0;
}

.jp-LinkedOutputView .maximizing-widget {
  height: 100%;
}

.jp-LinkedOutputView .jp-OutputArea-output:has(.maximizing-widget) {
  display: block;
}


:root {
  --jppype-foreground-color: var(--jp-content-font-color1);
  --jppype-dim-foreground-color: var(--jp-content-font-color2);

  --jppype-dim-background-color: var(--jp-layout-color3);
  --jppype-background-color: var(--jp-layout-color2);
  --jppype-highlight-background-color: var(--jp-layout-color1);
  --jppype-shadow-color: #000;
}