.ImageViewerWidget{
    height: 100%;
    min-height: 350px;

    width: 100%;
    min-width: 200px;
}

.jp-LinkedOutputView .ImageViewerWidget{
    height: 100%;
}

.ImageViewport{
    position: absolute;
    width: 100%;
    height: 100%;
    overflow: hidden;
    clip-path: polygon(0% 0%,100% 0%,100% 100%,0% 100%);
}

.ImageViewerWidget .hideForegroundLayers .foregroundLayer{
    display: none !important;
}

.jp-LinkedOutputView .ImageViewport:before {

}

.jp-LinkedOutputView .ImageViewport:before {
    content: '';
    position: absolute;
    left: 0;
    right: 0;
    width: calc(100% - 2px);
    height: calc(100% - 2px);
    z-index: 4;
    box-shadow: inset 0 0 2.5em -0.5em #1119;
    -webkit-box-shadow: inset 0 0 2.5em -0.5em #1119;
    -moz-box-shadow: inset 0 0 2.5em -0.5em #1119;
    border: var(--jp-layout-color2) 1px solid;
}

.ImageViewportOverlays{
    position: absolute;
    width: 100%;
    height: 100%;
}

.undraggable {
    -moz-user-select: none;
    -webkit-user-select: none;
    -ms-user-select: none;
    user-select: none;
    -webkit-user-drag: none;
    user-drag: none;
    -webkit-touch-callout: none;
}