// Copyright (c) Gabriel Lepetit-Aimon
// Distributed under the terms of the Modified BSD License.

// Import the CSS
import '../css/widget.css';

// Import ipywidgets
export { JView2D, JView2DModel } from './ipywidgets/JView2D';
