"use strict";(self.webpackChunkjppype=self.webpackChunkjppype||[]).push([[493],{9493:function(e,t,i){var r=this&&this.__createBinding||(Object.create?function(e,t,i,r){void 0===r&&(r=i);var n=Object.getOwnPropertyDescriptor(t,i);n&&!("get"in n?!t.__esModule:n.writable||n.configurable)||(n={enumerable:!0,get:function(){return t[i]}}),Object.defineProperty(e,r,n)}:function(e,t,i,r){void 0===r&&(r=i),e[r]=t[i]}),n=this&&this.__setModuleDefault||(Object.create?function(e,t){Object.defineProperty(e,"default",{enumerable:!0,value:t})}:function(e,t){e.default=t}),u=this&&this.__importStar||function(e){if(e&&e.__esModule)return e;var t={};if(null!=e)for(var i in e)"default"!==i&&Object.prototype.hasOwnProperty.call(e,i)&&r(t,e,i);return n(t,e),t};Object.defineProperty(t,"__esModule",{value:!0});const a=i(7464),o=i(2749),c=u(i(3216)),p=i(5064),l=i(1797),s={id:"jppype:plugin",requires:[a.IJupyterWidgetRegistry,o.ICommandPalette],activate:function(e,t,i){(0,p.initializeJApp)(e,i),t.registerWidget({name:l.MODULE_NAME,version:l.MODULE_VERSION,exports:c})},autoStart:!0};t.default=s}}]);